// Splay-  

//#include<stdlib.h>
//#include<stdio.h>
//#include<assert.h>
//#include<memory>
#include<algorithm>


class SplayTree 
{
public:
	typedef unsigned int value_type;
	struct Node
	{
		Node *p1, *p2;
		Node *parent;
		value_type value;
		Node(value_type value, Node *parent) : value(value), parent(parent), p1(0), p2(0)
		{}
	} *treeRoot;

	SplayTree() : treeRoot(0)
	{}
	~SplayTree()
	{
		while(treeRoot) 
		{
			Node *node = treeRoot;
			treeRoot = 0;
			std::swap(node->p1, treeRoot);  if (treeRoot) continue;
			std::swap(node->p2, treeRoot);  if (treeRoot) continue;
			treeRoot = node->parent;
			delete node;
		}
	}

	void Insert(value_type value)
	{
		Node *parent = 0;
		Node *node = treeRoot;
		while (node)
		{
			if (value == node->value) return;
			parent = node;
			node = (value < node->value) ? node->p1 : node->p2;
		}
		node = new Node(value, parent);
		if (!parent) treeRoot = node;
		else ((parent->value < node->value) ? parent->p2 : parent->p1) = node;
		Splay(node);
	}
	void Remove(value_type value)
	{
		Node *node = Find(value);
		if (!node) return;

		Splay(node);
		if (!node->p1) Replace(node, node->p2);
		else if (!node->p2) Replace(node, node->p1);
		else {
			Node *min = FindMin(node->p2);
			if (min->parent != node) {
				Replace(min, min->p2);
				min->p2 = node->p2;
				min->p2->parent = min;
			}
			Replace(node, min);
			min->p1 = node->p1;
			min->p1->parent = min;
		}
		delete node;
	}
	Node* Find(value_type value)
	{
		Node *node = treeRoot;
		while (node && value != node->value)
			node = value < node->value ? node->p1 : node->p2;
		if(node) Splay(node);
		return node;
	}
private:


	void FixParent(Node *node, Node *parent)
	{
		if (!node->parent) treeRoot = parent;
		else ((node == node->parent->p1) ? node->parent->p1 : node->parent->p2) = parent;
		node->parent = parent;
	}
	void Replace(Node *node, Node *n)
	{
		if (!node->parent) treeRoot = n;
		else ((node == node->parent->p1) ? node->parent->p1 : node->parent->p2) = n;
		if (n) n->parent = node->parent;
	}
	void RotateL(Node *node)
	{
		Node *p2 = node->p2;
		if (p2) {
			node->p2 = p2->p1;
			if (p2->p1)  p2->p1->parent = node;
			p2->parent = node->parent;
			p2->p1 = node;
		}
		FixParent(node, p2);
	}
	void RotateR(Node *node)
	{
		Node *p1 = node->p1;
		if (p1) {
			node->p1 = p1->p2;
			if (p1->p2) p1->p2->parent = node;
			p1->parent = node->parent;
			p1->p2 = node;
		}
		FixParent(node, p1);
	}
	void Splay(Node *node)
	{
		while (node->parent)
		{
			Node *p1 = node->parent->p1;
			Node *p2 = node->parent->p2;
			Node *pp = node->parent->parent;
			if (!pp) {
				if (p1 == node) RotateR(node->parent);
				else RotateL(node->parent);
			}
			else if (pp->p2 == node->parent) {
				if (p2 == node) RotateL(node->parent->parent);
				else RotateR(node->parent);
				RotateL(node->parent);
			}
			else if (pp->p1 == node->parent) {
				if (p1 == node) RotateR(node->parent->parent);
				else RotateL(node->parent);
				RotateR(node->parent);
			}
		}
	}
	Node *FindMin(Node *node)
	{
		while (node->p1) node = node->p1;
		return node;
	}
};



int main()
{
	SplayTree tree;
    for (int p=0; p<1000; p++) 
	{
		tree.Insert(rand() % 100);
		tree.Remove(rand() % 100);
	}
    return 0;
}

